'''# 网易云音乐 APIs

## 编写
对 EAPI (非网页端所用) 请求，抓包后可将数据体 `params=` 后内容丢进 `utils/eapidumper.py` 直接生成相应 Python 函数。
对于 Weapi (网页端，小程序所用) 请求，可在 `core.js` 内下断点后手动编写

编写的所有的 API 过程应遵循下列格式：

    @[API 请求类型] # 可根据抓到包的 URL 判断
    @[API 限制条件] # 可选
    @[API 预、后处理] # 可选
    def [API 名]([API 参数]):
        """[Google 风格多行注释]"""
        return '[API URL]',{[API 参数]}

对于需要 `checkToken` 参数的 API,请使用 `Crypto.checkToken()` 计算 - 该参数时效性约为 1 月
'''
import base64
from logging import getLogger
from time import time
from requests.models import Response
from .. import GetCurrentSession,SetCurrentSession
from .. import logger
from ..utils import Crypto
import json,urllib.parse

class LoginRequiredException(Exception):pass
class LoginFailedException(Exception):pass
# region Base models export
LOGIN_REQUIRED = LoginRequiredException('Function needs you to be logged in')

def parse(url):return urllib.parse.urlparse(url)

def BaseWrapper(requestFunc):
    '''Base wrapper for crypto requests'''
    def apiWrapper(apiFunc):
        '''apiFunc -> rtype : (url,plain,[method])'''
        def wrapper(*a,**k):
            ret       = apiFunc(*a,**k)
            url,plain = ret[:2]
            method    = ret[-1] if ret[-1] in ['POST','GET'] else 'POST'        
            rsp = requestFunc(url,plain,method)                        
            try:
                payload = rsp.text if isinstance(rsp,Response) else rsp
                payload = payload.decode() if not isinstance(payload,str) else payload
                return json.loads(payload.strip('\x10')) # weird caveat - they padded plaintext responses as well,though they always end with \x10(16)
            except json.JSONDecodeError:
                return rsp
        return wrapper
    return apiWrapper

def LoginRequiredApi(func):
    '''限制条件 - API 需要事先登录'''
    def wrapper(*a,**k):
        if not GetCurrentSession().login_info['success']:raise LOGIN_REQUIRED
        return func(*a,**k)
    return wrapper

def UserIDBasedApi(func):
    '''限制条件 - API 第一参数为用户 ID，而该参数可留 0 而指代已登录的用户 ID'''
    def wrapper(user_id=0,*a,**k):
        if user_id == 0 and GetCurrentSession().login_info['success']:
            user_id = GetCurrentSession().login_info['content']['account']['id']
        elif user_id == 0:
            raise LOGIN_REQUIRED
        return func(user_id,*a,**k)
    return wrapper

def EapiEncipered(func):
    '''后处理 - 函数值有 Eapi 加密 - 解密并返回原文'''
    def wrapper(*a,**k):
        payload = func(*a,**k)
        try:            
            return Crypto.EapiDecrypt(payload).decode()
        except:
            return payload
    return wrapper        

@BaseWrapper
def WeapiCryptoRequest(url,plain,method):
    '''请求类型 - Weapi - 适用于 网页端、小程序、手机端部分 APIs'''
    payload = json.dumps({**plain,'csrf_token':GetCurrentSession().csrf_token})
    return GetCurrentSession().request(method,
        url.replace('/api/','/weapi/'),
        params={'csrf_token':GetCurrentSession().csrf_token},
        data={**Crypto.WeapiCrypto(payload)},
    )
# region Port of `Binaryify/NeteaseCloudMusicApi`
@BaseWrapper
def LapiCryptoRequest(url,plain,method):
    '''请求类型 - Linux API - 适用于 Linux 客户端部分 APIs'''
    payload = {
        'method':method,
        'url':GetCurrentSession().HOST + url,
        'params':plain
    }
    payload = json.dumps(payload)
    return GetCurrentSession().request(method,
        '/api/linux/forward',
        headers={
            **GetCurrentSession().headers, 
            'User-Agent':GetCurrentSession().UA_LINUX_API
        },
        data={**Crypto.LinuxCrypto(payload)}
    )

@BaseWrapper
@EapiEncipered
def EapiCryptoRequest(url,plain,method):
    '''请求类型 - Eapi - 适用于 新版客户端绝大部分 APIs'''
    cookies = {
        **GetCurrentSession().CONFIG_EAPI,  
        'requestId':f'{int(time() * 1000)}_0233',      
        '__csrf':GetCurrentSession().csrf_token,
        **GetCurrentSession().cookies,
    }
    payload = {
        **plain,    
        'header':json.dumps(cookies)
    }
    request = GetCurrentSession().request(method,
        url,
        headers={
            **GetCurrentSession().headers, 
            'User-Agent':GetCurrentSession().UA_EAPI,
            'Referer':None
        },
        cookies=cookies,
        data={**Crypto.EapiCrypto(parse(url).path.replace('/eapi/','/api/'),json.dumps(payload))}
    )
    return request.content
# endregion
# endregion
from . import miniprograms,album,cloud,cloudsearch,login,playlist,track,user,video


'''Implementation of some of NE's crypto functions'''
import base64
from Crypto.Cipher import AES
from Crypto.Random import random
from hashlib import md5
from . import checkToken
class RSAPublicKey():
    def __init__(self, n, e):
        '''RSA pubkey pubkey & modulus pair,values are interpeted as integers'''
        self.n = int(n, 16)
        self.e = int(e, 16)
# region Constant values
WEAPI_AES_KEY    = "0CoJUm6Qyw8W8jud" # cbc
WEAPI_AES_IV     = "0102030405060708" # cbc
WEAPI_RSA_PUBKEY = RSAPublicKey(
    "00e0b509f6259df8642dbc35662901477df22677ec152b5ff68ace615bb7b725152b3ab17a876aea8a5aa76d2e417629ec4ee341f56135fccf695280104e0312ecbda92557c93870114af6c9d05c4f7f0c3685b7a46bee255932575cce10b424d813cfe4875d3e82047b97ddef52741d546b8e289dc6935b3ece0462db0a22b8e7",
    "10001" # textbook rsa without padding
)
LINUXAPI_AES_KEY = "rFgB&h#%2?^eDg:Q" # ecb
EAPI_DIGEST_SALT = "nobody%(url)suse%(text)smd5forencrypt"
EAPI_DATA_SALT   = "%(url)s-36cd479b6b5-%(text)s-36cd479b6b5-%(digest)s"
EAPI_AES_KEY     = "e82ckenh8dichen8" # ecb
BASE62           = 'PJArHa0dpwhvMNYqKnTbitWfEmosQ9527ZBx46IXUgOzD81VuSFyckLRljG3eC'
# endregion

class Crypto():
    '''The cryptography toolkit'''
    # region Base cryptograhpy methods
    # region Utility
    checkToken = checkToken
    @staticmethod
    def RandomString(len, chars=BASE62):
        '''Generates random string of `len` chars within a selected number of chars'''
        return ''.join([random.choice(chars) for i in range(0, len)])
    @staticmethod
    def HexDigest(data : bytearray):
        '''Digests a `bytearray` to a hex string'''
        return ''.join([hex(d)[2:].zfill(2) for d in data])
    @staticmethod
    def HexCompose(hexstr : str):
        '''Composes a hex string back to a `bytearray`'''
        return bytearray([int(hexstr[i:i+2],16) for i in range(0,len(hexstr),2)])
    @staticmethod
    def HashDigest(text):
        '''Digests 128 bit md5 hash'''
        HASH = md5(text.encode('utf-8'))
        return HASH.digest()        
    @staticmethod
    def HashHexDigest(text):
        '''Digests 128 bit md5 hash,then digest it as a hexstring'''
        return Crypto.HexDigest( Crypto.HashDigest(text) )
    # endregion
    # region Cryptos
    @staticmethod
    def AESEncrypt(
        data:str, key:str, mode=AES.MODE_ECB , iv=''       
    ):
        '''Basic AES encipher function'''
        def pad(data,blocksize=AES.block_size):return data + (blocksize - len(data) % blocksize) * chr(blocksize - len(data) % blocksize)        
        if mode in [AES.MODE_EAX,AES.MODE_ECB]:
            encryptor = AES.new(key.encode(), mode)
        else:    
            encryptor = AES.new(key.encode(), mode, iv.encode())
        encrypt_aes = encryptor.encrypt(str.encode(pad(data)))
        return encrypt_aes
    @staticmethod
    def AESDecrypt(
        data:str,key:str,mode=AES.MODE_ECB,iv=''
    ):
        '''Basic AES decipher funtion'''
        def unpad(data):
            pad = data[-1]
            if not pad in range(0,AES.block_size):return data # data isn't padded
            return data[:-pad]
        if mode in [AES.MODE_EAX,AES.MODE_ECB]:
            decryptor = AES.new(key.encode(), mode)
        else:    
            decryptor = AES.new(key.encode(), mode, iv.encode())
        decrypted_aes = decryptor.decrypt(data)
        return unpad(decrypted_aes)
    @staticmethod
    def RSAEncrypt(data:str, pubkey: RSAPublicKey, reverse=True):
        '''Signle-block textbook RSA encrpytion (c ≡ n ^ e % N),encodes text to hexstring first'''
        n = data if not reverse else reversed(data)        
        n, e, N = int(''.join(n).encode('utf-8').hex(), 16), pubkey.e, pubkey.n
        r = pow(n,e,N) # n ** e % N - modular exponetiation
        return Crypto.HexCompose(hex(r)[2:].zfill(256))
    # endregion
    # endregion
    # region Netease crypto
    '''source:core.js thingy'''
    @staticmethod
    def WeapiCrypto(params, aes_key2=None):
        '''Used in web & PC client'''
        aes_key2 = aes_key2 or Crypto.RandomString(16)
        params = str(params)
        # 1st go,encrypt the text with aes_key and aes_iv
        params = str(base64.encodebytes(Crypto.AESEncrypt(
            data=params, key=WEAPI_AES_KEY, iv=WEAPI_AES_IV,mode=AES.MODE_CBC)), encoding='utf-8')
        # 2nd go,encrypt the ENCRYPTED text again,with the 2nd key and aes_iv
        params = str(base64.encodebytes(Crypto.AESEncrypt(
            data=params, key=aes_key2, iv=WEAPI_AES_IV,mode=AES.MODE_CBC)), encoding='utf-8')
        # 3rd go,generate RSA encrypted encSecKey
        encSecKey = Crypto.HexDigest(Crypto.RSAEncrypt(aes_key2, WEAPI_RSA_PUBKEY))        
        return {
            'params': params,
            'encSecKey': encSecKey
        }
    '''source:
        - decompilation of `libpoison.so` : https://juejin.im/post/6844903586879520775
        - Binaryify/NeteaseCloudMusicApi  : https://github.com/Binaryify/NeteaseCloudMusicApi/blob/master/util/crypto.js'''
    @staticmethod
    def LinuxCrypto(params):
        '''Used in desktop linux clients'''
        params = str(params)
        # Single-pass AES-128 ECB crypto
        return {
            'eparams':Crypto.HexDigest(Crypto.AESEncrypt(params,key=LINUXAPI_AES_KEY,mode=AES.MODE_ECB))
        }
    @staticmethod
    def EapiCrypto(url,params):
        '''Used in mobile and PC clients'''
        url,params = str(url),str(params)
        digest = Crypto.HashHexDigest(EAPI_DIGEST_SALT % {'url':url,'text':params})
        params = EAPI_DATA_SALT % ({'url':url,'text':params,'digest':digest})
        return {
            'params':Crypto.HexDigest(Crypto.AESEncrypt(params,key=EAPI_AES_KEY,mode=AES.MODE_ECB))
        }
    @staticmethod
    def EapiDecrypt(cipher):
        '''Used in mobile clients'''
        cipher = bytearray(cipher) if isinstance(cipher,str) else cipher
        return Crypto.AESDecrypt(cipher,EAPI_AES_KEY,mode=AES.MODE_ECB) if cipher else cipher
    # endregion